;;; -*- mode: lisp; coding: utf-8 -*-
;;; 
;;; Copyright (c) 2007 Masayuki Onjo <onjo@lispuser.net>
;;; 
;;; Redistribution and use in source and binary forms, with or without
;;; modification, are permitted provided that the following conditions
;;; are met:
;;; 
;;;   * Redistributions of source code must retain the above copyright
;;;     notice, this list of conditions and the following disclaimer.
;;; 
;;;   * Redistributions in binary form must reproduce the above
;;;     copyright notice, this list of conditions and the following
;;;     disclaimer in the documentation and/or other materials
;;;     provided with the distribution.
;;; 
;;; THIS SOFTWARE IS PROVIDED BY THE AUTHOR 'AS IS' AND ANY EXPRESSED
;;; OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
;;; WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
;;; ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY
;;; DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
;;; DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
;;; GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
;;; INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
;;; WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
;;; NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
;;; SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

(defpackage :mecab (:use :cl :cffi :jp) (:export :mecab-new :mecab-destroy :wakati :sparse-tostr :sparse-tonode))
(in-package :mecab)

;;; Package configuration

(defvar *mecab-encoding* #+mswindows (jp:make-encoding :sjis)
                         #-mswindows (jp:make-encoding :utf8))

;;; Basic Utility Macro

(defmacro with-collector ((var &optional (collect-fn-name 'collect) (flush-fn-name 'flush)) &body body)
  (let (($var var)
        ($ptr (gensym)))
    `(let ((,$var nil)
           (,$ptr nil))
       (flet ((,collect-fn-name (item)
                (if (null ,$var)
                    (setf ,$var (cons item nil)
                          ,$ptr ,$var)
                    (setf (cdr ,$ptr) (cons item nil)
                          ,$ptr (cdr ,$ptr))))
              (,flush-fn-name ()
                (prog1 ,$var (setf ,$var nil))))
         ,@body)
       ,$var)))

(defmacro with-collectors (lst &body body)
  (if (null lst)
      `(progn ,@body)
      `(with-collector ,(first lst)
         (with-collectors ,(rest lst) ,@body))))

;;; Define Library

(define-foreign-library libmecab
  (:unix "libmecab.so")
  (:windows "c:/Program Files/MeCab/bin/libmecab.dll"))

(use-foreign-library libmecab)

;;; C-level API

(defcfun ("mecab_version" %mecab-version) :string)
(defcfun ("mecab_new2" %mecab-new2) :pointer
  (arg :string))
(defcfun ("mecab_destroy" %mecab-destroy) :void
  (m :pointer))
(defcfun ("mecab_sparse_tostr" %mecab-sparse-tostr) :pointer
  (m :pointer)
  (arg :string))
(defcfun ("mecab_sparse_tonode" %mecab-sparse-tonode) :pointer
  (m :pointer)
  (arg :string))

(defcstruct mecab-path
  (rnode :pointer)
  (rnext :pointer)
  (lnode :pointer)
  (lnext :pointer)
  (cost  :int)
  (prob  :float))

(defcstruct mecab-node
  (prev :pointer) ;  struct mecab_node_t  *prev;
  (next :pointer) ; struct mecab_node_t  *next;
  (enext :pointer) ; struct mecab_node_t  *enext;
  (bnext :pointer) ; struct mecab_node_t  *bnext;
  (rpath :pointer) ; struct mecab_path_t  *rpath;
  (lpath :pointer) ; struct mecab_path_t  *lpath;
  (begin_node_list :pointer) ; struct mecab_node_t **begin_node_list;
  (end_node_list :pointer)   ; struct mecab_node_t **end_node_list;
  (surface :string)
  (feature :string)
  (id :int) ; unsigned int          id;
  (length :short) ; unsigned short        length; /* length of morph */
  (rlength :short) ; unsigned short        rlength; /* real length of morph(include white space before the morph) */
  (rcAttr :short) ; unsigned short        rcAttr;
  (lcAttr :short) ; unsigned short        lcAttr;
  (posid :short) ; unsigned short        posid;
  (char_type :short) ; unsigned char         char_type;
  (stat :short) ; unsigned char         stat;
  (isbest :char) ; unsigned char         isbest;
  (sentence_length :uint) ; unsigned int          sentence_length; /* it is avaialbe only when BOS node */
  (alpha :float ) ; float                 alpha;
  (beta :float) ; float                 beta;
  (prob :float) ; float                 prob;
  (wcost :short) ; short                 wcost;
  (cost :short) ; long                  cost;
)

;;; Lisp-level Internal API

(defparameter *default-mecab* nil)

(defmacro with-foreign-vector ((name vec) &body body &aux ($v (gensym)))
  `(let ((,$v ,vec))
     (with-foreign-object (,name :uchar (length ,$v))
       (dotimes (i (length ,$v))
	 (setf (mem-ref ,name :uchar i) (aref ,$v i)))
       ,@body)))

(defun mecab-new (&optional (args ""))
  (with-foreign-vector (str (jp:encode args *mecab-encoding* :null-terminate t))
    (%mecab-new2 str)))

(defun mecab-destroy (mecab)
  (when mecab
    (%mecab-destroy mecab)))

;;; Lisp-level External API

(defun wakati (text &aux ptr (vec (typecase text ((vector (unsigned-byte 8)) text) (string (jp:encode text *mecab-encoding* :null-terminate t)))))
  (let ((mecab (mecab-new "-O wakati")))
    (unwind-protect
         (with-foreign-vector (cstr vec)
           (setf ptr (%mecab-sparse-tostr mecab cstr))
           (unless (null-pointer-p ptr)
             (with-collectors ((result)
                               (chunk collect-chunk flush-chunk))
                 (loop for i from 0
                       for c = (mem-aref ptr :uchar i)
                       until (= c 0) do
                       (if (= c (char-code #\space))
			   (collect (jp:decode (coerce (flush-chunk) '(vector (unsigned-byte 8))) *mecab-encoding*))
                           (collect-chunk c))))))
      (mecab-destroy mecab))))

(defun sparse-tostr (text &key (mecab *default-mecab*)
                          &aux ptr (vec (typecase text ((vector (unsigned-byte 8)) text) (string (jp:encode text *mecab-encoding* :null-terminate t)))))
  (unless mecab
    (setf *default-mecab* (mecab-new)
          mecab           *default-mecab*))
  (with-foreign-vector (cstr vec)
    (setf ptr (%mecab-sparse-tostr mecab cstr))
    (unless (null-pointer-p ptr)
      (loop for i from 0
         until (= (mem-aref ptr :uchar i) 0)
         collect (mem-aref ptr :uchar i) into result
         finally (let* ((vec (coerce result '(vector (unsigned-byte 8))))
                        (str (jp:decode vec *mecab-encoding*)))
                   (return
                     (with-collector (lst)
                       (loop with beg = 0
                             with tab = 0
                             with end = 0
                             for i from 0
                             for c across str
                             when (char= c #\Tab) do
                             (setf tab i)
                             when (char= c #\Newline) do
                             (setf end i)
                             (when (/= beg tab)
                               (collect (list (subseq str beg tab)
                                              :prop
                                              (ppcre:split "," (subseq str (1+ tab) end)))))
                              (setf beg (1+ i)
                                    tab beg)))))))))

(defun sparse-tonode (text &key (mecab *default-mecab*)
                          &aux ptr)
  (unless mecab
    (setf *default-mecab* (mecab-new)
          mecab           *default-mecab*))
  (with-foreign-vector (cstr (jp:encode text *mecab-encoding* :null-terminate t))
    (setf ptr (%mecab-sparse-tonode mecab cstr))
    (if (null-pointer-p ptr) NIL ptr)))

;;; Sample

(defun text-to-words (text)
  (let ((nodes (sparse-tostr text)))
    (loop for node in nodes collect (car node))))

(defun text-to-kana (text)
  (let ((nodes (sparse-tostr text)))
    (loop for node in nodes
          for sujo = (getf (cdr node) :prop)
          for yomi = (or (nth 8 sujo) (car node))
          when yomi
          do (princ yomi)
          finally (terpri))))
